unit columns;

{ column printing

  NOTE Columns are stored as arrays.
  By convention these are 1-based and never have more elements
  than are acually used

  eg for a four column definition use
  ASetOfCols : array[1..4] of TCol

}

interface

  uses classes,useful,printers,sysUtils,graphics;

  type

  TCol = record           // column data
      Left   : Integer;
      Width  : Integer;
      Decimal : integer;  // pixels from RH edge of decimal point
      Head   : String;    // heading text
      V      : Integer;   // furthest down page this col has got
      TopV   : Integer;   // Place vertical lines should reach up to
  end;

  TColAlign = (caLeft,caWrap,caCentre,caRight,caDecimal);

  THstyle = (hsShrink,         // shrink font size to go into column
             hsTilt,           // allow tilting if won't fit horizontally
             hsBold,           // print in bold
             hsNoText,         // don't print any text

             hsLowerBorder,    // mark border at base of text
             hsUpperBorder,    // mark border at top of text (default)
             hsRule,           // single rule
             hsMultiRule,      // nice multi-ruled underlines
             hsBoxes,          // draw a box around each heading
             hsOneBox          // draw a single box around all titles
             );                // heading styles

  THStyles = Set of THStyle;

  Function CentraliseColumns(var c:array of Tcol; const LM,RM:integer):integer;
  procedure ColourColumn(var c:array of Tcol; const Colno,VertPos,Depth:integer; const Colour:TColor);
  procedure ColVerticals(const c: array of TCol;const Width:integer);
  function DrawBreakLine(var c:array of TCol):integer;
  procedure DrawColBorder(const c: array of TCol;const Width:integer);
  procedure PrintColHead(const V:Integer; var H:TCol; const Styles: THstyles;
                  const FontName:string; const FontSize:integer);
  function PrintHeadings(const V:integer; var c:array of TCol;
                       const Styles: THstyles;
                       const FontName:string;
                       const FontSize:integer):integer;
  function PrintInCol(var c:array of Tcol;
                    const Colno,VertPos:integer;
                    const S:string;
                    const Align:Tcolalign):integer;
  Function SetColumnMargins(var c:array of tcol;const LM:integer):integer;
  procedure SetColumnTitles(var c:array of tcol; const t: array of string);
  procedure SetColumnWidths(var c:array of tcol; const w: array of integer);
  procedure VerticalColLine(const Bottom,Horiz,Width:integer;C:Tcol);
  function VMaximum(c:array of TCol):integer;
implementation




Function DrawBreakLine(var c:array of TCol):integer;
{ draw a line underneath the columns at the furthest
  point reached down the page
  return how how far down the page we get}
var
 i,V,HL,HR : Integer;
begin
  V := Vmaximum(c)+pix(1);
  with printer.canvas do begin
    pen.width := 1;// (pix(25) div 100);
    pen.style := psSolid;
    for i := low(c) to high(c) do begin
      HL := c[i].left;
      HR := HL +c[i].width;
      moveto(HL+(pix(1)div 2),V);
      lineto(HR-(pix(1)div 2),V);
    end;
  end;
  result := V;
  c[0].v := result+2;    // make sure we won't go back by setting 1st col to this amount
end;


procedure DrawColBorder(const c: array of TCol;const Width:integer);
{ draw a surrounding border to the complete table of columns}
var
  L,T,R,B : Integer;
begin
  L := c[low(c)].left;
  R := c[high(c)].left+c[high(c)].width;
  T := c[low(c)].topv;
  B := vmaximum(c);
  with printer.canvas do begin
    pen.width := width;
    moveto(L,T);
    lineto(R,T);
    lineto(R,B);
    lineto(L,B);
    lineto(L,T);
  end;
end;



procedure PrintColHead(const V:Integer;
                         var H:TCol;
                   const Styles: THstyles;
                const FontName:string;
                const FontSize:integer);
{ print a single column heading at the vertical position specified}
var
  i,T,B,L,R : Integer;
  S : String;
  angd : integer;
  angr : single;
  Y,W:integer;
begin
  T := V;
  B := T + pix(6);
  L := H.left;
  R := L + H.width;
  S := H.Head;
  H.topV := V;       // default position of top border
  H.v := B+pix(2);   // mark (default) last vertical print place

  setfont(FontName,FontSize);

  with printer.canvas do begin

    // ************* TEXT **************************************************
    if not (hsNoText in styles) then begin

      // possibly bold--------------------------------------------------------
      if (hsbold in styles) then font.style := [fsBold] else font.style := [];

      // possibly shrink text-------------------------------------------------
      if (hsShrink in styles) then begin
        while textwidth(S)>(R-L) do begin
          font.size := font.size - 1;
          if font.size < 6 then break;
        end;
      end;

      // possibly tilt--------------------------------------------------------
      if (hsTilt in styles) then begin
        angd := 60;                     // degrees
        angr := (angd * 3.14159 )/ 180;  // radians
        i := textheight(S);
        Y := B - trunc(i * cos(angr));
        canvastextoutangle(printer.canvas,L,Y,(angd * 10),S);
        // this might have had the effect of moving the top of the
        // headings 'back up' the page as this is probably now
        // taller than the other headings.
        W := textheight(S);
        Y := Y - trunc(W * sin(angr));
        if H.topV > Y then H.topv := Y;
      end else begin
        // print plain text--------------------------------------------------
        pout(centralh(L,R,S),centralv(T,B,S),S);
      end;
    end;
    // we have finished printing the text


    //********* lines and boxes ***********************

    // simply set the top point of the column to the current base line------
    if (hsLowerBorder in Styles) then H.topv := H.v;

    // draw individual box---------------------------------------------------
    if (hsBoxes in styles) then begin
      with pen do begin
        style := pssolid;
        color := clBlack;
        width := 1;
      end;
      moveto(L+(pix(1) div 2),B);
      lineto(R-(pix(1) div 2),B);
      lineto(R-(pix(1) div 2),T);
      lineto(L+(pix(1) div 2),T);
      lineto(L+(pix(1) div 2),B);
    end;

    // draw single rule under heading-----------------------------------------
    if (hsRule in styles) then begin
      with pen do begin
        style := pssolid;
        color := clBlack;
        width := 1;
      end;
      for i := 0 to 5 do begin
        moveto(L+(pix(1) div 2),B+i);
        lineto(R-(pix(1) div 2),B+i);
      end;
    end;

    // nice multiple rulings------------------------------------------------
    if (hsMultirule in styles) then begin
      with pen do begin
        style := pssolid;
        color := clBlack;
        width := 1;
      end;
      for i := 0 to 5 do begin
        moveto(L+(pix(1) div 2),B+(i*3));
        lineto(R-(pix(1) div 2),B+(i*3));
      end;
      H.topV := B+pix(2);   // default position of top
    end;
  end;
end;

function PrintHeadings(const V:integer; var c:array of TCol;
                       const Styles: THstyles;
                       const FontName:string;
                       const FontSize:integer):integer;
{ Column headings starting at vertical position V
  Return furthest point reached down the page
  Fintname and size as for SETFONT}
var
  L,R,T,B,VP,i : integer;
begin
  T := V + pix(2);
  for i := low(c) to high(c) do begin
    printcolhead(T,c[i],Styles,FontName,FontSize);
  end;
  // watch out in case we had to tilt any headings
  // this could push the top of the column back up the page
  // only applicable if the
  //************* where to set the top border position?**********
  if (hsTilt in styles) then begin
    if not (hsLowerBorder in Styles) then begin
      VP := 99999;
      for i := low(c) to high(c) do if VP> C[i].topV then VP := c[i].topv;
      for i := low(c) to high(c) do c[i].topv := VP;
    end;
  end;

  result := Vmaximum(c)+2;

  // possibly draw a big box round the headings
  if (hsOneBox in Styles) then begin
    with printer.canvas do begin
      pen.width := 1;
      B := result;
      L := c[low(c)].left;
      R := c[high(c)].left +c[high(c)].width;
      moveto(L,T);
      lineto(R,T);
      lineto(R,B);
      lineto(L,B);
      lineto(L,T);
    end;
  end;
end;

Function CentraliseColumns(var c:array of Tcol; const LM,RM:integer):integer;
{ adjust column positions so that it is central
  see also setcolumnmargins
  Return maximum RH extent }
var
  r,wrequired,wprovided : integer;
begin
  r := setcolumnmargins(c,lm);
  wrequired := 1+r-lm;                    // widths
  wProvided := 1+RM-LM;
  result := setcolumnmargins(c,LM+((wProvided-wRequired) div 2));
end;

Function SetColumnMargins(var c:array of tcol;const LM:integer):integer;
{ * USE AFTER SETTING WIDTHS *
  Given the left margin of the first column we can
  work out the left margin of all columns because
  we just increment each time by the width.
  Return the right hand position of the last column
  ie the overall width of the columns
  NB also sets the vertical extents to 0}
var
  i : integer;
begin
  c[low(c)].left := LM;
  c[low(c)].v := 0;
  for i := low(c) to high(c)-1 do begin
    c[i+1].left := c[i].left + c[i].width;
    c[i+1].v := 0;
  end;
  result := c[high(c)].left + c[high(c)].width;
end;


procedure SetColumnTitles(var c:array of tcol;const t: array of string);
{ set all the titles in one hit.
  example SetColumnTitles(Cols,['Item','Each','Qty','Amount']);
  NB also sets the vertical extents to 0}
var
  i : integer;
begin
  for i := low(c) to high(c) do begin
    c[i].v := 0;
    try
      c[i].head := t[i];
    except
      raise exception.createfmt('Problem assigning title to column %d',[i+1]);
      // i+1 because when array passed passed to proc it is forced to zero-base
    end;
  end;
end;

procedure SetColumnWidths(var c:array of tcol;const w: array of integer);
{ set all the widths in one hit.
  example SetColumnWidths(Cols,[12,14,pix(55),12,12*aNumber]);}
var
  i : integer;
begin
  for i := low(c) to high(c) do begin
    try
      c[i].width := w[i];
      c[i].decimal := w[i] div 4;  // some default for decimal point
    except
      raise exception.createfmt('Problem assigning width to column %d',[i+1]);
      // i+1 because when array passed passed to proc it is forced to zero-base
    end;
  end;
end;

procedure ColVerticals(const c: array of TCol;const Width:integer);
{ draw all the verticals for this table from the lowest point to the
  tops of the headings}
var
  b,i : integer;
begin
  b := vmaximum(c);
  for i := low(c) to high(c) do begin
    verticalcolline(b,c[i].left,width,c[i]);
  end;
  verticalcolline(b,c[high(c)].left+c[high(c)].width,width,c[high(c)]);
end;



procedure VerticalColLine(const Bottom,Horiz,Width:integer;C:Tcol);
{ draw a vertical line at the horizontal position specified
  from the bottom position to the top of the column.

  * Negative width of pen gives a dotted line

  (The top of the column will have been determined by the heading style
  when it was printed}
begin
  with printer.canvas do begin
    pen.width := abs(width);
    if width < 0 then pen.style := psdot;
    moveto(Horiz,Bottom);
    lineto(Horiz,C.TopV);
    pen.style := pssolid;
  end;
end;


function  VMaximum(C:array of TCol):integer;
{ Return the furthest point down the page
  reached in any of the columns
  Update all the columns to match the furthest}
var
  i : integer;
begin
  result := 0;
  for i := low(C) to high(C) do begin
    if C[i].v > result then result := C[i].v;
  end;
  for i := low(C) to high(C) do begin
    C[i].v := result;
  end;

end;


procedure ColourColumn(var c:array of Tcol; const Colno,VertPos,Depth:integer; const Colour:TColor);
{ Fill the specified column with the specified colour
  This is a bit of a kludge because we have to call this BEFORE writing text
  as I can't get copyrect to work as advertised.
  Depth is the number of pixels height we expect to fill }
var
  L,R:integer;
begin
  // get the basic limits we're working to
  L := c[Colno-1].left;
  R := L + c[colno-1].width;

  // get out the paints
  with printer.canvas do begin
    brush.style := bssolid;
    brush.color := colour;
    fillrect(rect(L,Vertpos,R,Vertpos+depth-1));
    brush.style := bsclear;
  end;
end;

function PrintInCol(var c:array of Tcol;
                    const Colno,VertPos:integer;
                    const S:string;
                    const Align:Tcolalign):integer;
{ print the string S into the specified column
  Style of printing depends on the AlignArgument
     caLeft       Left justify - try to shrink text if it won't fit
     caWrap       Wrap
     caCentre     Centralise - shrink if it won't fit
     caRight      Right Justify - shrink if it won't fit
     caDecimal    Decimal align
  if the Vertical Pos is <0 then use the furthest point so far in this column
  Return the new vertical page location  }

var
  L,R,V,halfmm,savefontsize : integer;
  {sp : integer;
  F : string;
  Fract:boolean;}
begin
  // get the basic limits we're working to
  halfmm := pix(-10) div 20;
  L := c[Colno-1].left;
  R := L + c[colno-1].width;
  L := l + halfmm;
  R := R - halfmm;
  V := VertPos;
  if v < 0 then V := c[colno-1].V;

  with printer.canvas do begin

    // possibly shrink the text to fit
    savefontsize := font.size;
    if align in [caLeft,caCentre,caRight,caDecimal] then begin
      while textwidth(S)>c[colno-1].width do begin
        if font.size < 7 then break;
        font.size := font.size -1;
      end;
    end;

    case Align of
      caLeft   : pline(L,V,S);
      caCentre : pline(centralH(L,R,S),V,S);
      caRight  : pline(R - textwidth(S),V,S);
      caDecimal: begin
                   pdecimal(R-c[colno-1].decimal,V,S);
                   V := V + textheight(S);
                 end;
      caWrap   : pwrap(L,R,V,S);
      (*caWrapFractions : begin
           //***** this is used by Stockval when printing securities with
           //***** fractions in them.  We only give this special treatment
           //***** to strings that end "....<space><digits><slash><digits>p"
           //***** eg "... 10/19p"
           fract := rightstr(S,1)='p';
           if fract then begin
             sp = posminus(' ',S);
             fract = (sp <> 0);
             if fract then begin
               F := copy(s,sp+1,(length(S)-sp)-1);  //F = n/n
               fract = pos('/',F)>0;
             end;
           end;
           if not fract then begin
             pwrap(L,R,V,S);              // warp as we would normally
           end else begin
             pwrap(L,R,V,copy(s,1,sp));
             TO BE COMPLETED

        *)




    end;
    c[colno-1].V := V;
    result := V;

    if align in [caLeft,caCentre,caRight,caDecimal] then font.size := savefontsize;

  end;
end;

end.
